// This file was developed by Thomas Müller <thomas94@gmx.net>.
// It is published under the BSD 3-Clause License within the LICENSE file.

#include <tev/FalseColor.h>
#include <tev/ImageCanvas.h>
#include <tev/ThreadPool.h>

#include <nanogui/theme.h>

#define STB_IMAGE_WRITE_IMPLEMENTATION
#include <stb_image_write.h>

#include <numeric>

using namespace Eigen;
using namespace nanogui;
using namespace std;

TEV_NAMESPACE_BEGIN

ImageCanvas::ImageCanvas(nanogui::Widget* parent, float pixelRatio)
: GLCanvas(parent), mPixelRatio(pixelRatio) {
    setDrawBorder(false);
}

bool ImageCanvas::scrollEvent(const Vector2i& p, const Vector2f& rel) {
    if (GLCanvas::scrollEvent(p, rel)) {
        return true;
    }

    scale(rel.y(), p.cast<float>());
    return true;
}

void ImageCanvas::drawGL() {
    if (!mImage) {
        mShader.draw(
            2.0f * mSize.cast<float>().cwiseInverse() / mPixelRatio,
            Vector2f::Constant(20)
        );
        return;
    }

    if (!mReference) {
        mShader.draw(
            2.0f * mSize.cast<float>().cwiseInverse() / mPixelRatio,
            Vector2f::Constant(20),
            mImage->texture(getChannels(*mImage)),
            // The uber shader operates in [-1, 1] coordinates and requires the _inserve_
            // image transform to obtain texture coordinates in [0, 1]-space.
            transform(mImage.get()).inverse().matrix(),
            mExposure,
            mOffset,
            mTonemap
        );
        return;
    }

    mShader.draw(
        2.0f * mSize.cast<float>().cwiseInverse() / mPixelRatio,
        Vector2f::Constant(20),
        mImage->texture(getChannels(*mImage)),
        // The uber shader operates in [-1, 1] coordinates and requires the _inserve_
        // image transform to obtain texture coordinates in [0, 1]-space.
        transform(mImage.get()).inverse().matrix(),
        mReference->texture(getChannels(*mReference)),
        transform(mReference.get()).inverse().matrix(),
        mExposure,
        mOffset,
        mTonemap,
        mMetric
    );
}

void ImageCanvas::draw(NVGcontext *ctx) {
    GLCanvas::draw(ctx);

    if (mImage) {
        auto texToNano = textureToNanogui(mImage.get());
        auto nanoToTex = texToNano.inverse();

        Vector2f pixelSize = texToNano * Vector2f::Ones() - texToNano * Vector2f::Zero();

        Vector2f topLeft = (nanoToTex * Vector2f::Zero());
        Vector2f bottomRight = (nanoToTex * mSize.cast<float>());

        Vector2i startIndices = Vector2i{
            static_cast<int>(floor(topLeft.x())),
            static_cast<int>(floor(topLeft.y())),
        };

        Vector2i endIndices = Vector2i{
            static_cast<int>(ceil(bottomRight.x())),
            static_cast<int>(ceil(bottomRight.y())),
        };

        if (pixelSize.x() > 50) {
            float fontSize = pixelSize.x() / 6;
            float fontAlpha = min(1.0f, (pixelSize.x() - 50) / 30);

            vector<string> channels = getChannels(*mImage);
            // Remove duplicates
            channels.erase(unique(begin(channels), end(channels)), end(channels));

            vector<Color> colors;
            for (const auto& channel : channels) {
                colors.emplace_back(Channel::color(channel));
            }

            nvgFontSize(ctx, fontSize);
            nvgFontFace(ctx, "sans");
            nvgTextAlign(ctx, NVG_ALIGN_CENTER | NVG_ALIGN_MIDDLE);

            Vector2i cur;
            vector<float> values;
            for (cur.y() = startIndices.y(); cur.y() < endIndices.y(); ++cur.y()) {
                for (cur.x() = startIndices.x(); cur.x() < endIndices.x(); ++cur.x()) {
                    Vector2i nano = (texToNano * (cur.cast<float>() + Vector2f::Constant(0.5f))).cast<int>();
                    getValuesAtNanoPos(nano, values);

                    TEV_ASSERT(values.size() >= colors.size(), "Can not have more values than channels.");

                    for (size_t i = 0; i < colors.size(); ++i) {
                        string str = tfm::format("%.4f", values[i]);
                        Vector2f pos{
                            mPos.x() + nano.x(),
                            mPos.y() + nano.y() + (i - 0.5f * (values.size() - 1)) * fontSize,
                        };

                        // First draw a shadow such that the font will be visible on white background.
                        nvgFontBlur(ctx, 2);
                        nvgFillColor(ctx, Color(0.0f, fontAlpha));
                        nvgText(ctx, pos.x() + 1, pos.y() + 1, str.c_str(), nullptr);

                        // Actual text.
                        nvgFontBlur(ctx, 0);
                        Color col = colors[i];
                        nvgFillColor(ctx, Color(col.r(), col.g(), col.b(), fontAlpha));
                        nvgText(ctx, pos.x(), pos.y(), str.c_str(), nullptr);
                    }
                }
            }
        }
    }

    // If we're not in fullscreen mode draw an inner drop shadow. (adapted from nanogui::Window)
    if (mPos.x() != 0) {
        int ds = mTheme->mWindowDropShadowSize, cr = mTheme->mWindowCornerRadius;
        NVGpaint shadowPaint = nvgBoxGradient(
            ctx, mPos.x(), mPos.y(), mSize.x(), mSize.y(), cr * 2, ds * 2,
            mTheme->mTransparent, mTheme->mDropShadow
        );

        nvgSave(ctx);
        nvgResetScissor(ctx);
        nvgBeginPath(ctx);
        nvgRect(ctx, mPos.x(), mPos.y(), mSize.x(), mSize.y());
        nvgRoundedRect(ctx, mPos.x() + ds, mPos.y() + ds, mSize.x() - 2 * ds, mSize.y() - 2 * ds, cr);
        nvgPathWinding(ctx, NVG_HOLE);
        nvgFillPaint(ctx, shadowPaint);
        nvgFill(ctx);
        nvgRestore(ctx);
    }
}

void ImageCanvas::translate(const Vector2f& amount) {
    mTransform = Translation2f(amount) * mTransform;
}

void ImageCanvas::scale(float amount, const Vector2f& origin) {
    float scaleFactor = pow(1.1f, amount);

    // Use the current cursor position as the origin to scale around.
    Vector2f offset = -(origin - position().cast<float>()) + 0.5f * mSize.cast<float>();
    auto scaleTransform =
        Translation2f(-offset) *
        Scaling(scaleFactor) *
        Translation2f(offset);

    mTransform = scaleTransform * mTransform;
}

float ImageCanvas::applyExposureAndOffset(float value) {
    return pow(2.0f, mExposure) * value + mOffset;
}

vector<string> ImageCanvas::getChannels(const Image& image) const {
    vector<vector<string>> groups = {
        { "R", "G", "B" },
        { "r", "g", "b" },
        { "X", "Y", "Z" },
        { "x", "y", "z" },
        { "U", "V" },
        { "u", "v" },
        { "Z" },
        { "z" },
    };

    string layerPrefix = mRequestedLayer.empty() ? "" : (mRequestedLayer + ".");

    vector<string> result;
    for (const auto& group : groups) {
        for (size_t i = 0; i < group.size(); ++i) {
            const auto& name = layerPrefix + group[i];
            if (image.hasChannel(name)) {
                result.emplace_back(name);
            }
        }

        if (!result.empty()) {
            break;
        }
    }

    string alphaChannelName = layerPrefix + "A";

    // No channels match the given groups; fall back to the first 3 channels.
    if (result.empty()) {
        const auto& channelNames = image.channelsInLayer(mRequestedLayer);
        for (const auto& name : channelNames) {
            if (name != alphaChannelName) {
                result.emplace_back(name);
            }

            if (result.size() >= 3) {
                break;
            }
        }
    }

    // If we found just 1 channel, let's display is as grayscale by duplicating it twice.
    if (result.size() == 1) {
        result.push_back(result[0]);
        result.push_back(result[0]);
    }

    // If there is an alpha layer, use it
    if (image.hasChannel(alphaChannelName)) {
        result.emplace_back(alphaChannelName);
    }

    return result;
}

Vector2i ImageCanvas::getImageCoords(const Image& image, Vector2i mousePos) {
    Vector2f imagePos = textureToNanogui(&image).inverse() * mousePos.cast<float>();
    return {
        static_cast<int>(floor(imagePos.x())),
        static_cast<int>(floor(imagePos.y())),
    };
}

void ImageCanvas::getValuesAtNanoPos(Vector2i mousePos, vector<float>& result) {
    result.clear();
    if (!mImage) {
        return;
    }

    Vector2i imageCoords = getImageCoords(*mImage, mousePos);
    const auto& channels = getChannels(*mImage);

    for (const auto& channel : channels) {
        result.push_back(mImage->channel(channel)->eval(imageCoords));
    }

    // Subtract reference if it exists.
    if (mReference) {
        Vector2i referenceCoords = getImageCoords(*mReference, mousePos);
        const auto& referenceChannels = getChannels(*mReference);
        for (size_t i = 0; i < result.size(); ++i) {
            float reference = i < referenceChannels.size() ?
                mReference->channel(referenceChannels[i])->eval(referenceCoords) :
                0.0f;

            result[i] = applyMetric(result[i], reference);
        }
    }
}

Vector3f ImageCanvas::applyTonemap(const Vector3f& value) const {
    Vector3f result;
    switch (mTonemap) {
        case ETonemap::SRGB:
            {
                static const auto toSRGB = [](float linear) {
                    if (linear < 0.0031308f) {
                        return 12.92f * linear;
                    } else {
                        return 1.055f * pow(linear, 0.41666f) - 0.055f;
                    }
                };

                result = {toSRGB(value.x()), toSRGB(value.y()), toSRGB(value.z())};
                break;
            }
        case ETonemap::Gamma:
            {
                static const float gamma = 2.2f;
                result = {pow(value.x(), 1 / gamma), pow(value.y(), 1 / gamma), pow(value.z(), 1 / gamma)};
                break;
            }
        case ETonemap::FalseColor:
            {
                static const auto falseColor = [](float linear) {
                    static const auto& fcd = falseColorData();
                    int start = 4 * clamp((int)(linear * (fcd.size() / 4)), 0, (int)fcd.size() / 4 - 1);
                    return Vector3f{fcd[start], fcd[start + 1], fcd[start + 2]};
                };

                result = falseColor(log2(value.mean()) / 10 + 0.5f);
                break;
            }
        case ETonemap::PositiveNegative:
            {
                result = {-2.0f * value.cwiseMin(Vector3f::Zero()).mean(), 2.0f * value.cwiseMax(Vector3f::Zero()).mean(), 0.0f};
                break;
            }
        default:
            throw runtime_error{"Invalid tonemap selected."};
    }

    return result.cwiseMax(Vector3f::Zero()).cwiseMin(Vector3f::Ones());
}

float ImageCanvas::applyMetric(float image, float reference) const {
    float diff = image - reference;
    switch (mMetric) {
        case EMetric::Error:                 return diff;
        case EMetric::AbsoluteError:         return abs(diff);
        case EMetric::SquaredError:          return diff * diff;
        case EMetric::RelativeAbsoluteError: return abs(diff) / (reference + 0.01f);
        case EMetric::RelativeSquaredError:  return diff * diff / (reference * reference + 0.01f);
        default:
            throw runtime_error{"Invalid metric selected."};
    }
}

void ImageCanvas::fitImageToScreen(const Image& image) {
    Vector2f nanoguiImageSize = image.size().cast<float>() / mPixelRatio;
    mTransform = Scaling(mSize.cast<float>().cwiseQuotient(nanoguiImageSize).minCoeff());
}

void ImageCanvas::resetTransform() {
    mTransform = Affine2f::Identity();
}

void ImageCanvas::saveImage(const string& filename) {
    if (!mImage) {
        return;
    }

    const auto& channels = channelsFromDisplayedImage();
    Vector2i imageSize = channels.front().size();
    size_t numPixels = (size_t)imageSize.x() * imageSize.y();

    TEV_ASSERT(channels.size() <= 4, "Can not save an image with more than 4 channels.");

    if (channels.empty()) {
        return;
    }

    cout << "Saving currently displayed image as " << filename << "... " << flush;
    auto start = chrono::system_clock::now();

    // Flatten channels into single array
    vector<float> floatData(4 * channels.front().data().size(), 0);

    ThreadPool pool;
    pool.parallelFor(0, channels.size(), [&channels, &floatData](size_t i) {
        const auto& channelData = channels[i].data();
        for (size_t j = 0; j < channelData.size(); ++j) {
            floatData[j * 4 + i] = channelData[j];
        }
    });

    // Manually set alpha channel to 1 if the image does not have one.
    if (channels.size() < 4) {
        for (size_t i = 0; i < numPixels; ++i) {
            floatData[i * 4 + 3] = 1;
        }
    }

    string lowerFilename = toLower(filename);
    if (endsWith(lowerFilename, ".hdr")) {
        // Store as HDR image.
        stbi_write_hdr(filename.c_str(), imageSize.x(), imageSize.y(), 4, floatData.data());
    } else {
        // Store as LDR image.
        vector<char> byteData(floatData.size());
        pool.parallelFor(0, numPixels, [&](size_t i) {
            size_t start = 4 * i;
            Vector3f value = applyTonemap({
                applyExposureAndOffset(floatData[start]),
                applyExposureAndOffset(floatData[start + 1]),
                applyExposureAndOffset(floatData[start + 2]),
            });
            for (int j = 0; j < 3; ++j) {
                floatData[start + j] = value[j];
            }
            for (int j = 0; j < 4; ++j) {
                byteData[start + j] = (char)(floatData[start + j] * 255);
            }
        });

        if (endsWith(lowerFilename, ".jpg") || endsWith(lowerFilename, ".jpeg")) {
            stbi_write_jpg(filename.c_str(), imageSize.x(), imageSize.y(), 4, byteData.data(), 100);
        } else if (endsWith(lowerFilename, ".png")) {
            stbi_write_png(filename.c_str(), imageSize.x(), imageSize.y(), 4, byteData.data(), 0);
        } else if (endsWith(lowerFilename, ".bmp")) {
            stbi_write_bmp(filename.c_str(), imageSize.x(), imageSize.y(), 4, byteData.data());
        } else if (endsWith(lowerFilename, ".tga")) {
            stbi_write_tga(filename.c_str(), imageSize.x(), imageSize.y(), 4, byteData.data());
        } else {
            throw invalid_argument{tfm::format("Image '%s' has unknown format.", filename)};
        }
    }

    auto end = chrono::system_clock::now();
    chrono::duration<double> elapsedSeconds = end - start;

    cout << tfm::format("done after %.3f seconds.\n", elapsedSeconds.count());
}

shared_ptr<Lazy<shared_ptr<CanvasStatistics>>> ImageCanvas::canvasStatistics() {
    if (!mImage) {
        return nullptr;
    }

    string channels = join(getChannels(*mImage), ",");
    string key = mReference ?
        tfm::format("%s-%s-%s-%d", mImage->filename(), channels, mReference->filename(), mMetric) :
        tfm::format("%s-%s", mImage->filename(), channels);

    auto iter = mMeanValues.find(key);
    if (iter != end(mMeanValues)) {
        return iter->second;
    }

    mMeanValues.insert(make_pair(key, make_shared<Lazy<shared_ptr<CanvasStatistics>>>([this]() {
        return computeCanvasStatistics();
    }, &mMeanValueThreadPool)));

    auto val = mMeanValues.at(key);
    val->computeAsync();
    return val;
}

vector<Channel> ImageCanvas::channelsFromDisplayedImage() const {
    if (!mImage) {
        return {};
    }

    vector<Channel> result;
    const auto& channelNames = getChannels(*mImage);
    for (size_t i = 0; i < channelNames.size(); ++i) {
        result.emplace_back(toUpper(Channel::tail(channelNames[i])), mImage->size());
    }

    if (!mReference) {
        ThreadPool pool;
        pool.parallelFor(0, channelNames.size(), [&](size_t i) {
            const auto* chan = mImage->channel(channelNames[i]);
            const auto& channelData = chan->data();

            for (size_t j = 0; j < channelData.size(); ++j) {
                result[i].data()[j] = channelData[j];
            }
        });
    } else {
        Vector2i size = mImage->size();
        Vector2i offset = (mReference->size() - size) / 2;
        const auto& referenceChannels = getChannels(*mReference);

        ThreadPool pool;
        pool.parallelFor(0, channelNames.size(), [&](size_t i) {
            const auto* chan = mImage->channel(channelNames[i]);
            bool isAlpha = result[i].name() == "A";

            if (i < referenceChannels.size()) {
                const Channel* referenceChan = mReference->channel(referenceChannels[i]);
                if (isAlpha) {
                    for (int y = 0; y < size.y(); ++y) {
                        for (int x = 0; x < size.x(); ++x) {
                            result[i].at({x, y}) = 0.5f * (
                                chan->eval({x, y}) +
                                referenceChan->eval({x + offset.x(), y + offset.y()})
                            );
                        }
                    }
                } else {
                    for (int y = 0; y < size.y(); ++y) {
                        for (int x = 0; x < size.x(); ++x) {
                            result[i].at({x, y}) = applyMetric(
                                chan->eval({x, y}),
                                referenceChan->eval({x + offset.x(), y + offset.y()})
                            );
                        }
                    }
                }
            } else {
                if (isAlpha) {
                    for (int y = 0; y < size.y(); ++y) {
                        for (int x = 0; x < size.x(); ++x) {
                            result[i].at({x, y}) = chan->eval({x, y});
                        }
                    }
                } else {
                    for (int y = 0; y < size.y(); ++y) {
                        for (int x = 0; x < size.x(); ++x) {
                            result[i].at({x, y}) = applyMetric(chan->eval({x, y}), 0);
                        }
                    }
                }
            }
        });
    }

    return result;
}

shared_ptr<CanvasStatistics> ImageCanvas::computeCanvasStatistics() const {
    const auto& flattened = channelsFromDisplayedImage();

    float mean = 0;
    float maximum = -numeric_limits<float>::infinity();
    float minimum = numeric_limits<float>::infinity();

    size_t nChannels = 0;

    for (const auto& channel : flattened) {
        if (channel.name() == "A") {
            continue;
        }

        auto channelSize = channel.size();
        size_t numElements = (size_t)channelSize.x() * channelSize.y();
        for (size_t i = 0; i < numElements; ++i) {
            float val = channel.eval(i);
            mean += val;
            maximum = max(maximum, val);
            minimum = min(minimum, val);
        }

        if (numElements > 0) {
            mean /= numElements;
        }

        ++nChannels;
    }

    auto result = make_shared<CanvasStatistics>();

    result->mean = nChannels > 0 ? (mean / nChannels) : 0;
    result->maximum = maximum;
    result->minimum = minimum;

    // Now that we know the maximum and minimum value we can define our histogram bin size.
    static const int NUM_BINS = 400;
    result->histogram = MatrixXf::Zero(NUM_BINS, nChannels);

    // We're going to draw our histogram in log space.
    auto symmetricLog2 = [](float val) {
        static const float addition = 0.001f;
        static const float smallest = log2(addition);
        return val > 0 ? (log2(val + addition) - smallest) : -(log2(-val + addition) - smallest);
    };

    float minLog2 = symmetricLog2(minimum);
    float maxLog2 = symmetricLog2(maximum);

    auto valToBin = [&](float val) {
        return clamp((int)(NUM_BINS * (symmetricLog2(val) - minLog2) / (maxLog2 - minLog2)), 0, NUM_BINS - 1);
    };

    size_t iChannel = 0;
    for (const auto& channel : flattened) {
        if (channel.name() == "A") {
            continue;
        }

        auto channelSize = channel.size();
        size_t numElements = (size_t)channelSize.x() * channelSize.y();
        for (size_t i = 0; i < numElements; ++i) {
            int bin = valToBin(channel.eval(i));
            result->histogram(bin, iChannel) += 1;
        }

        ++iChannel;
    }

    result->histogram /= result->histogram.maxCoeff() * 1.3f;
    result->histogramZero = valToBin(0);
    return result;
}

Vector2f ImageCanvas::pixelOffset(const Vector2i& size) const {
    // Translate by half of a pixel to avoid pixel boundaries aligning perfectly with texels.
    // The translation only needs to happen for axes with even resolution. Odd-resolution
    // axes are implicitly shifted by half a pixel due to the centering operation.
    // Additionally, add 0.1111111 such that our final position is almost never 0
    // modulo our pixel ratio, which again avoids aligned pixel boundaries with texels.
    return Vector2f{
        size.x() % 2 == 0 ?  0.5f : 0.0f,
        size.y() % 2 == 0 ? -0.5f : 0.0f,
    } + Vector2f::Constant(0.1111111f);
}

Transform<float, 2, 2> ImageCanvas::transform(const Image* image) {
    if (!image) {
        return Transform<float, 2, 0>::Identity();
    }

    // Center image, scale to pixel space, translate to desired position,
    // then rescale to the [-1, 1] square for drawing.
    return
        Scaling(2.0f / mSize.x(), -2.0f / mSize.y()) *
        mTransform *
        Scaling(1.0f / mPixelRatio) *
        Translation2f(pixelOffset(image->size())) *
        Scaling(image->size().cast<float>()) *
        Translation2f(Vector2f::Constant(-0.5f));
}

Transform<float, 2, 2> ImageCanvas::textureToNanogui(const Image* image) {
    if (!image) {
        return Transform<float, 2, 0>::Identity();
    }

    // Move origin to centre of image, scale pixels, apply our transform, move origin back to top-left.
    return
        Translation2f(0.5f * mSize.cast<float>()) *
        mTransform *
        Scaling(1.0f / mPixelRatio) *
        Translation2f(-0.5f * image->size().cast<float>() + pixelOffset(image->size()));
}

TEV_NAMESPACE_END
